///////////////////////////////////////////////////////
// Apple 2 Keyboard emulation
//  by Jesus Arias (2023)
//  PS2 keyboard. Mapping via ROM

`define BRAM_PROMK

module keybps2(
input clk,			// CPU clock
input kclk,			// PS2 clock
input kdat,			// PS2 data
output [6:0]keycode,	// output key code
output keyvalid,		// flag for valid keycode (1 cycle) 
output apple1,
output apple2,
output shift,
output [7:0]switches	// Reset key pressed

);

////////////////////////////////////////////////////
// PS2 serial receiver
////////////////////////////////////////////////////

reg [10:0]kreg=11'h7FF;	// shift register, 11-bit
reg krck=1;				// sampled keyboard clock
always @(posedge clk) krck<=kclk;	

reg kock=1;			// old clock sample (for edge detection)
reg rel=0;					// key released (if last scancode was 0xF0)
wire newsc=~kreg[0];		// zero start bit => scancode received
wire [7:0]scancode=kreg[8:1];

wire fall=(~krck)&kock;	// falling clock
reg [11:0]toutc=0;		// timeout counter
always @(posedge clk) toutc<= fall? 0: toutc+1;
wire tout=(toutc==12'hFFF);

// shift reg.gets filled with ones after a key is received
// or on timeouts
always @(posedge clk) begin
	kock<=krck;
	kreg<=(newsc|tout) ?11'h7FF : ( fall ? {kdat,kreg[10:1]} : kreg);
	if (newsc) rel<=(scancode==8'hF0);
end

/// Modiffiers
reg shift=0;
reg control=0;
reg alt=0;
reg caps=1;
reg swhide=0;
reg apple1=0;
reg apple2=0;
reg apple3=0;
reg res=0;
reg ini=0;
always @(posedge clk) begin
	if(newsc) begin
		// Modiffiers
		if ((scancode==8'h12)|(scancode==8'h59)) shift<=~rel;
		if ((scancode==8'h14)) control<=~rel;
		if ((scancode==8'h11)) alt<=~rel;
		if ((scancode==8'h58)&(~rel)) caps<=~caps;
		// some switches too
		if ((scancode==8'h4)&(~rel)) swhide<=~swhide;	// F3: VLEDs
		if ((scancode==8'h1))  apple1<=~rel; // F9
		if ((scancode==8'h9))  apple2<=~rel; // F10
		if ((scancode==8'h78)) ini<=~rel; // F11: Init (reset all)
		if ((scancode==8'h7))  res<=~rel; // F12: Apple II reset
	end
end

wire swreset= res;
wire swinit = ini;

/// Pulsed switches
wire swcolor=(scancode==8'hc)&newsc&(~rel);			// F4: colorsel
//wire swdsk1=(scancode==8'h5)&newsc&(~rel);			// F2: dsk1
//wire swdsk2=(scancode==8'h6)&newsc&(~rel);			// F3: dsk2
assign switches = {1'b0,swhide,swinit,caps,2'b00,swcolor,swreset};


////////////////////////////////////////////////////
// scancode to keyboard array mapping (async ROM)
////////////////////////////////////////////////////
//  $00 -  $7F Base map (no modiffiers)
//  $80 -  $FF Caps lock map
// $100 - $17F Shift map
// $180 - $1FF Alt map

wire [1:0]gs = alt ? 3 : ( shift ? 2 : caps);
`ifdef BRAM_PROMK
genrom #(
	.INITFILE("keyrom.hex"),
	.AW(9),
	.DW(7)
	) prom6 (.clk(clk), .addr({gs,scancode[6:0]}), .data_out(romq));
wire [6:0]romq;
reg nnsc=0;
always @(posedge clk) nnsc <= newsc & (~rel);
wire kval = nnsc & (romq!=0);
`else
reg [6:0]keyrom[0:511];
initial begin
  $readmemh("keyrom.hex", keyrom);
end
wire [6:0]romq=keyrom[{gs,scancode[6:0]}];
wire kval = newsc & (~rel) & (romq!=0);
`endif

reg keyvalid=0;
always @(posedge clk) keyvalid<=kval;
reg [6:0]keycode=0;
always @(posedge clk) 
	if (kval) keycode <= {romq[6]&(~control),romq[5]&(~control),romq[4:0]};

endmodule

