///////////////////////////////////////////////////////
// Apple IIe computer recreated
//  by Jesus Arias (2023)
//

`define SOFTCARD	// Microsoft Z80 card for CPM
`define SOUNDS	// floppy noises
`define FLOPPYTON

`include "cpu6502.v"
`include "video.v"
`include "keyboard.v"
`include "wozmachine.v"

`ifdef FLOPPYTON
`define Z80
`include "floppyton.v"
`endif

`ifdef SOUNDS
`include "sound.v"
`endif

`ifdef SOFTCARD
`define Z80
`endif

`ifdef Z80
`include "tv80s.v"
`include "tv80_core.v"
`include "tv80_alu.v"
`include "tv80_reg.v"
`include "tv80_mcode.v"
`endif

module SYSTEM (
	input clk2x,		// 25 MHz
	input reset,

	output xwe,			// external RAM: WE
	output xoe,			// external RAM: OE
	output [15:0]xa,	// external RAM: address
	output [15:0]xdo,	// external RAM: data output
	input  [15:0]xdi,	// external RAM: data input
	output xbhe,		// external RAM: high byte
	output xble,		// external RAM: low byte 
	// Serial port
	output txd,		
	input rxd,
	// VGA signals
	output hsyn,		
	output vsyn,
	output [11:0]video,	// RGB444 (4 bits per color)
	// SPI bus (To Flash and SD card)
	output sck,
	output mosi,
	output ss1,		// Flash /CS
	output ss2,		// SD card /CS
	input  miso,
	// Audio
	output	pwmout,
	// Keyboard
	input	kclk,
	input	kdat
	, output [6:0]debug
	
);

//-- divide clock by 2
reg clk=0;
always @(posedge clk2x) clk<=~clk; 

/////////////////// RESET /////////////////
wire greset = reset | swinit;	// global reset
wire areset = greset|kreset;

/////////////////// CLKs ////////////////////

reg [3:0]ckcnt=0;		// Divider: 1/12

reg ck1M=1;				// CPU clock
`ifdef SOFTCARD
wire cktop=(ckcnt==(zsel?4'd5:4'd11));
wire clk65=cclk|zsel;
wire clkz80=cclk|(~zsel);
`else
wire cktop=(ckcnt==4'd11);
wire clk65=cclk;
`endif
always @(posedge clk) begin
	ckcnt<= cktop ? 0 : ckcnt+1;
	if (cktop) ck1M<=0;
	if ((~ck1M)&(~vrd)) ck1M<=1; // Rising edge delayed if this is a video read cycle
end

//wire cclk = clk | vrd | ck1M;
wire cclke = vrd | (ck1M&(~fastclk));
wire cclk = clk | cclke | ftwait;
wire clk2M = (cclke & (ckcnt!=4'd6)) | clk;

/////////////////////////////////////////////////////////////////////
/////////////////////////////// CPUs ////////////////////////////////
/////////////////////////////////////////////////////////////////////
// 							(Primary CPU)
// 			Arlet Ottens 6502 core with synchronous memory 

wire [15:0]pca;
wire [7:0]cdi = csboot ? bootdo : cdi_reg;
wire [7:0]pcdo;
wire pwe;
wire irq;

// Beware: all signals active high!
cpu6502 cpu65( .AB(pca),.DO(pcdo),.WE(pwe),.DI(cdi),.clk(clk65),.reset(areset),
	.IRQ(1'b0), .NMI(1'b0), .RDY(1'b1) );

// registered data input
wire [7:0]cdi_d=iocs ? iodi : ((introm|lcma16|(~acxxx))?xdi8:8'hff);
reg [7:0]cdi_reg; 
always @(posedge cclk) cdi_reg<=cdi_d;

/////////// Z80 Softcard stuff /////////////
`ifdef SOFTCARD
	// CPU select Flip-Flop
	reg zsel=0;		// Z80 selected
	always @(posedge clk or posedge areset) 
		if (areset) zsel<=0; 
		else if ((~cclke)&(ca[15:8]==8'hC4)&(~boot)&we) zsel<=~zsel;

	//////////////////////////////////////////////////////////////////
	// 		Guy Hutchison's TV80 core with asynchronous memory
	//////////////////////////////////////////////////////////////////
	wire reset_n,wait_n,int_n,nmi_n,busrq_n,cen;
	wire [15:0]za;
	wire [15:0]zca;
	wire [7:0]zcdo;
	wire m1_n,mreq_n,iorq_n,rd_n,wr_n,rfsh_n,halt_n,busak_n;

	assign reset_n=~(areset);
	assign wait_n=1'b1;

	assign nmi_n=1'b1;
	assign busrq_n=1'b1;
	assign cen=1'b1;

	tv80s #(.Mode(1)) Z80cpu(//AUTOARG
	  // Outputs
	  m1_n, mreq_n, iorq_n, rd_n, wr_n, rfsh_n, halt_n, busak_n, za, zcdo,
	  // Inputs
	  reset_n, clkz80, wait_n, int_n, nmi_n, busrq_n, cdi_d, cen
	  );
	  
	// Softcard address translation
	reg [3:0]zham; // modiffied high address bits, not reg
	always @*
	 case (za[15:12])
	 4'hb: zham<=4'hd;
	 4'hc: zham<=4'he;
	 4'hd: zham<=4'hf;
	 4'he: zham<=4'hc;
	 default: zham<=za[15:12]+1;
	 endcase

	///// BUS mux /////
	wire [15:0]ca = zsel? {zham,za[11:0]} : pca;
	wire [7:0]cdo = zsel? zcdo : pcdo;
	wire we = zsel? (~mreq_n)&(~wr_n) : pwe;
`else
	wire zsel=0;
	wire [15:0]ca = pca;
	wire [7:0]cdo = pcdo;
	wire we= pwe;
`endif
	wire rw = ~we;

/////////// address decoding & MMU ////////////////
wire iocs  = (ca[15:8]==8'hC0);
wire romcs = ca[15]&ca[14]&(~iocs);
reg csboot =0;	// boot memory delayed 1 cycle
always @(posedge cclk) csboot<= ca[15]&ca[14]&ca[13]&boot;

wire acxxx=(ca[15:12]==4'hc);			// Slot
wire ac6xx=acxxx&(ca[11:8]==4'h6);
wire sc6xx=ac6xx&(~cxrom);
wire ac3xx=acxxx&(ca[11:8]==4'h3);
wire acfff=(ca[15:0]==16'hcfff);
wire ac8xx=(ca[15:11]==5'b11001);
reg intc8rom=0;
always @(posedge cclk or posedge areset)
	if (areset) intc8rom<=0; 
	else if (acfff) intc8rom<=0;
	else if (ac3xx & (~c3xrom)) intc8rom<=1;

wire introm=(acxxx & cxrom)|(ac3xx & (~c3xrom))|(ac8xx & intc8rom) | sc6xx;

////////////////////////////// VIDEO ///////////////////////////////
wire vrd;
wire [16:0]va;
wire hsyn;
wire vsyn,vblank;
wire [11:0]a2video;

wire vpage=page2 & (~_80store);

video video0( .clk2x(clk2x), .clk(clk), .hsyn(hsyn), .vsyn(vsyn), .vblank(vblank), .video(a2video),
			.text(text), .mix(mixed), .page2(vpage), .hires(hires),
			.rd(vrd), .a(va), .d(xdi8), .colorsel(color), 
			._80col(_80col), .an3(an3), .altchar(altchar),
			.enleds(~swhide), .vleds(vleds)
 ); 

wire atext= (ca[15:10]==6'b000001);
wire ahires=(ca[15:13]==3'b001);

reg [1:0]color=0;
always @(posedge cclk) if (swcolor) color<=color+1;

wire [7:0]vleds = {3'b000,zsel,dskwrrq,drvon&drvsel,drvon&(~drvsel),swcaps};

///////////////////////////////////////////////////
// external RAM 

wire azp=(ca[15:9]==0);
wire amain = (~azp)&(~(ca[15]&ca[14]));
wire en80v = ((hires&ahires&_80store)|(atext&_80store)) ;
wire en80m = ((auxwr&we)|(auxrd&(~we))) & ((~en80v)&amain);
wire en80zp= azp & auxzp;
wire en80zph=((lcwren&we)|(lcrden&(~we))) & apaged & auxzp;

wire ma16=lcma16 | (en80v&page2)| en80m| en80zp | en80zph;
wire noxwrite = en80zph | en80m | en80zp;


wire [16:0]ma = vrd ? {va} : {ma16,ca[15:13],lcma12^ca[12],ca[11],
	ca[10]&(~sc6xx),ca[9]&(~sc6xx), ca[8:0]};

assign xdo = {cdo,cdo};
assign xa = ma[16:1];
assign xble = ma[0];
assign xbhe = ~ma[0];

assign xoe = ~( vrd | rw ); 

wire wprot = romcs&(~boot)&(~ma16);
assign xwe = noxwrite| wprot | (~( (~vrd)& we & (~iocs) & (~cclk) ));
wire [7:0]xdi8 = ma[0] ? xdi[15:8] : xdi[7:0];  

///////////////////////// INTERNAL boot ROM//////////////////////////////
wire [7:0]bootdo;
genrom #(
	.INITFILE("boot.hex"),
	.AW(9),
	.DW(8)
	) bootrom0 (.clk(cclk),.addr(ca[8:0]),.data_out(bootdo));

////////////////////////////////////////////////////////////////////
//                          PERIPHERALS
////////////////////////////////////////////////////////////////////

wire [7:0]iodi;	// data from peripherals to CPU

assign iodi =	(bootregcs	? {miso,1'b0,bootr}	: 8'hFF ) &
				(keycs		? {keyval,keycode[6:0]}	: 8'hFF ) &
				(keyrescs	? {kb7mux[ca[3:0]],keycode[6:0]}: 8'hFF ) &
				(c06xcs		? {serin[ca[2:0]],7'h7F}: 8'hFF ) &
				(slot6cs    ? slot6in               : 8'hFF );

// control input mux: 
wire [15:0]kb7mux={_80col, altchar,  hires,     page2,
				   mixed,     text, vblank,  _80store,
				   c3xrom,   auxzp,  cxrom,     auxwr,
				   auxrd,   lcrden, ~lcbank, keycode[7] };

// decoding

wire bootregcs = iocs & (ca[7:4]==4'h2) & boot;
wire alatchcs  = iocs & (ca[7:4]==4'h5);
wire spkcs     = iocs & (ca[7:4]==4'h3);
wire keycs     = iocs & (ca[7:4]==4'h0);
wire keyrescs  = iocs & (ca[7:4]==4'h1);
wire c06xcs    = iocs & (ca[7:4]==4'h6);
wire slot6cs   = iocs & (ca[7:4]==4'hE);
wire slot0cs   = iocs & (ca[7:4]==4'h8);

wire alatch2cs = iocs & (ca[7:4]==4'h0) & we;

// boot register
reg [5:0]bootr=6'b111111;
always @(posedge clk or posedge greset) 
	if (greset) bootr<=6'b111111; else if (bootregcs & we) bootr<=cdo[5:0];

wire boot = bootr[4];
wire fastclk = bootr[5];
assign mosi = bootr[0] | dskmosi;
assign sck  = bootr[1] | dsksck;
assign ss1  = bootr[2];
assign ss2  = bootr[3] & dskss2b;

// Addresable latch for video switches & annunciators

reg [1:0]alat01=0;	// bits 0 & 1 have no reset
always @(posedge cclk) 
	if (alatchcs&(~ca[3])&(~ca[2])) alat01[ca[1]]<=ca[0];
	
reg [7:0]alatch=0;	// other bits (bit 0 & 1 not used)
always @(posedge cclk or posedge areset) 
	if (areset) alatch<=0;
	else if (alatchcs) alatch[ca[3:1]]<=ca[0];
wire text =alat01[0];
wire mixed=alat01[1];
wire page2=alatch[2];
wire hires=alatch[3];
wire an0  =alatch[4];
wire an1  =alatch[5];
wire an2  =alatch[6];
wire an3  =alatch[7];

// Addresable latch for memory banking, 80col, etc
reg [7:0]alatch2=4'b0000;
always @(posedge cclk or posedge areset) 
	if (areset) alatch2<=0;
	else if (alatch2cs) alatch2[ca[3:1]]<=ca[0];
wire _80store=alatch2[0];
wire auxrd   =alatch2[1];
wire auxwr   =alatch2[2];
wire cxrom   =alatch2[3];
wire auxzp   =alatch2[4];
wire c3xrom  =alatch2[5];
wire _80col  =alatch2[6];
wire altchar =alatch2[7];

/////// Speaker ///////
reg speaker=0;
always @(posedge cclk) if (spkcs) speaker<=~speaker;

`ifdef SOUNDS
soundemu sound0 (
	.clk(clk), .phs(stepperph), .motor(drvon), .speaker(speaker), .pwm(pwmout)
);
`else
	assign pwmout =speaker;
`endif

///////////////// 16K upper RAM (Languaje Card) /////////////
reg [3:0]lcc=4'b1000;		// Control register	
wire drden=~(ca[1]^ca[0]);	// next read enable
wire dprewr=(~we)&ca[0];	// next pre-write enable
wire dwren=~(((dprewr&lcprewen)|(~lcc[3]))&ca[0]);	// next write enable
always @(posedge clk or posedge areset) 
	if (areset) lcc<=4'b1000; else if (slot0cs&(~cclke)) lcc<={dwren,dprewr,drden,ca[3]};
wire lcbank=lcc[0];		// current 4K bank for $D000
wire lcrden=lcc[1];		// 1: read RAM 0: read ROM
wire lcprewen=lcc[2];
wire lcwren=~lcc[3];	// 1: write RAM 0: no writes

wire apaged=ca[15]&ca[14]&(ca[13]|ca[12]); // CPU address range for paged memory

wire lcma16 = (apaged&lcrden&(~we))|(apaged&we&lcwren); // LC RAM in high mem for read and writes
wire lcma12 = (ma16&(~ca[13])&lcbank);	// toggle A12 for bank 2 in $D000 range


// Serial input (cassete, paddels, etc)
// bit 0: cassete in
// bits 1,2,3 : paddle buttons
// bits 4,5,6,7: NE555 ADCs for paddle pots
wire [7:0]serin = {5'b1111,~shift,apple2,apple1,1'b0};

/////////////////////////////////////////////
//////// Slot 6 (floppy disk emulation) /////
wire [7:0]slot6in;
wire [3:0]stepperph;
wire drvon, drvsel, dskwrprot, dskwrdata, dskrddata, dskwrrq;

wozmachine dskcntrl0 (.clk(cclk), .clk2x(clk2M | ftwait), .we(we), .cs(slot6cs), .a(ca[3:0]),
				.reset(areset), .d(cdo), .q(slot6in),
				.phs(stepperph), .drvon(drvon), .drvsel(drvsel), .wrprot(dskwrprot),
				.wrrq(dskwrrq), .wrdata(dskwrdata),	.rddata(dskrddata)
			);

//////////////////// Floppyton Computer //////////////////////
// Video sharing
wire ftvideo,ftrem, ftviden;
assign video[11:8]=ftviden? {ftvideo,ftvideo,ftvideo,ftvideo} :a2video[11:8];
assign video[7:4]=ftviden? {(ftrem)^ftvideo,ftvideo,ftvideo,ftvideo} :a2video[7:4];
assign video[3:0]=ftviden? ({1'b0,(~ftvideo)^ftrem,1'b0,(~ftvideo)^ftrem}) :a2video[3:0];

`ifdef FLOPPYTON
// SPI sharing
wire dsksck, dskmosi, dskss2b;
wire kmute;
wire ftwait;	// stop Apple II if asserted
floppyton floppytn( .clk(clk2x), .reset(greset), .hsyn(hsyn), .vsyn(vsyn),
		.video(ftvideo), .rem(ftrem), .viden(ftviden),
		.miso(miso), .sck(dsksck), .mosi(dskmosi), .cs2b(dskss2b),
		.kclk(kclk), .kdat(kdat), .grab(kmute), .drvon(drvon),
		.phs(stepperph), .drvsel(drvsel), .wrprot(dskwrprot),
		.wrrq(dskwrrq), .wrdata(dskwrdata),	.rddata(dskrddata),
		.waitwr(ftwait)
		,.debug(en_dma)
);

wire en_dma;
assign debug[0]=stepperph[0]|stepperph[2];
assign debug[1]=en_dma;
assign debug[2]=ftwait;

`else
wire kmute=0;
wire dsksck=0;
wire dskmosi=0;
wire dskss2b=1;
`endif
////////////// Keyboard /////////////
// micro-UART for keyboard emulation
// 115200 bps

reg [1:0]rx=0;

always @(posedge clk) rx<={rx[0],rxd};
reg [6:0]bauddiv=0;
always @(posedge clk)
	if ((rx[1]^rx[0])|(bauddiv==108)) bauddiv<=0;
	else bauddiv<=bauddiv+1;

wire sample = (bauddiv==55);

reg [9:0]rbuf=10'h3FF;
wire rxval = ~rbuf[0];
reg [7:0]keycode=0;
reg keyval=0;
always @(posedge clk) begin
	if (~rbuf[0]) rbuf<=10'h3FF;
	else if (sample) rbuf<={rx[0],rbuf[9:1]};
	
	if ((~rbuf[0])|(kvalps2&(~cclke))) keycode<=((~rbuf[0])? rbuf[8:1] :0 )|(kvalps2?{1'b0,keyps2}:0);
	if ((~rbuf[0])|(kvalps2&(~cclke))) keyval<=1;
	else if (keyrescs&(ca[3:0]==0)&(~cclke)) keyval<=0; 

end

assign txd = 1'b1; //rx[0];
//wire kreset=(~boot)&(keycode==18); // crtl-R reset

/// PS2 keyboard
wire [6:0]keyps2;
wire kvalps2in;
wire kvalps2 = kvalps2in & (~kmute);
wire [7:0]kdbg;
wire [7:0]switches;	// control signals
wire apple1,apple2,shift;
assign kreset = switches[0];
assign swcolor= switches[1];
assign swdsk1 = switches[2];
assign swdsk2 = switches[3];
assign swcaps = switches[4];
assign swinit = switches[5];
assign swhide = switches[6];

keybps2 ps2 ( .clk(clk | cclke), .kclk(kclk), .kdat(kdat),
	.keycode(keyps2), .keyvalid(kvalps2in), .apple1(apple1), .apple2(apple2), .shift(shift),
	.switches(switches)
);

endmodule


//----------------------------------------------------------------------------
//-- Generic ROM memory
//----------------------------------------------------------------------------

module genrom (        
    input clk,                      
    input wire [AW-1: 0] addr,      
    output reg [DW-1: 0] data_out   
);

parameter INITFILE = "character.hex"; // default hex file with contents (must be valid)
parameter AW = 9;	// number of address bits
parameter DW = 8;	// number of data bits

//localparam NPOS = 2 ** AW;
//localparam NPOS = 1<<AW;

  // memory
  reg [DW-1: 0] mem [0: (1<<AW)-1];

  // synchronous read (mandatory for BRAM instatiation)
  always @(posedge clk) begin
    data_out <= mem[addr];
  end

// initial values
initial begin
  $readmemh(INITFILE, mem);
end
endmodule

