///////////////////////////////////////////////////////
// Apple 2 VIDEO
//  by Jesus Arias (2023)
// VGA compatible video generation
// VESA 640x480 60Hz timing (more or less)

module video (
	input clk2x,	// clock input @25MHz
	input clk,		// clock input @12.5MHz
	input [7:0]d,	// data coming from memory
	output [16:0]a,	// memory address 
	output rd,		// read memory when 1
	output hsyn,	// horizontal sync.
	output vsyn,	// vertical sync.
	output vblank,	// vertical blank
	
	input text,		// Text (1) / Graph (0)
	input mix,		// 4 lines of text (1), only graph (0)
	input page2,	// video page
	input hires,	// Hires (1) / lores (0) 
	input _80col,	// Double res if 1 (80col)
	input an3,		// Double Hires if 1
	input altchar,  // Alternate character (lowercase)
	
	input [1:0]colorsel,	// NTSC color emulation when 1
	
	output [11:0]video,	// RGB444 video output (12 bits/pixel)
	input enleds,
	input [7:0]vleds
);

parameter LEDCOLOR=8'b1111_0001; 		// LED mask: 0=RED 1=Orange
parameter GREENCOLOR=12'b0011_1111_1000;	// Green Phosphor monitor
parameter AMBERCOLOR=12'b1111_1100_0001;	// Amber Phosphor monitor

///////////////// Horizontal counters ///////////////
reg hsyn=1;
reg hde=0;
// pixel counter (7 pixels / Character)
reg [2:0]pixcnt=0;
wire pixtc = pixcnt[2]&pixcnt[1];	// (pixcnt==6) (last pixels/character)
always @(posedge clk)
	if (pixtc | nl) pixcnt<=0;
	else pixcnt<=(pixcnt==6) ? 0 : pixcnt+1;

// Character counter
reg [5:0]chrcnt=0;
// New line flag (57*7 pixels + 1 pixel delay = 400 pixels /line)
wire nl=chrcnt[5]&chrcnt[4]&chrcnt[3]&chrcnt[0];	// Value 57

always @(posedge clk) begin
	if (nl) chrcnt<=0;
	else if (pixtc) chrcnt<=chrcnt+1;
end

// DE & HSYN
always @(posedge clk) begin
	hde<= nl ? 1 : ((pixtc&(chrcnt==39)) ? 0 : hde);
	if (pixtc) begin
		if (chrcnt==43) hsyn<=0;
		if (chrcnt==51) hsyn<=1;
	end
end

///////////////// Vertical counter ///////////////
reg [9:0]vc=523;	
reg vsyn=1;
reg vde=0;
reg l160=0;

always @(posedge hde) begin
	if (vc==10'd524) begin vc<=0; vde<=1; l160<=0; end 
	else vc<=vc+1;
	if (vc==10'd319) l160<=1;
	if (vc==10'd383) vde<=0;
	if (vc==10'd441) vsyn<=0;
	if (vc==10'd443) vsyn<=1;
end

assign vblank=vde;

wire texttime = text | (mix & l160);
wire dres = _80col&(texttime | (hires&(~an3)));
wire page = page2&(~dres);

///////////////// Data valid, Read //////////////
wire de=hde&vde;
wire rd1=(pixcnt==0);
wire rd2=(pixcnt==4)&dres;
assign rd=de&(rd1|rd2);

////////////// video address //////////////

wire [3:0]sum = {vc[8:7],vc[8:7]}+{1'b0,chrcnt[5:3]};
wire [9:0]txtva={vc[6:4],sum,chrcnt[2:0]};
	
// graphis/text switch
wire gr_en = (~text)& hires & (~(mix & l160));

// address generation
assign a[15:0]={ gr_en ? {1'b0,page,~page,vc[3:1]}: {4'b0000,page,~page}, txtva};
assign a[16]=dres&rd1;
	
//////////////// Character ROM /////////////////////
// flash timer
reg [4:0]osc=0;
always @(posedge vsyn) osc<= (osc==5'd30) ? 5'd1 : (osc+1);

reg lcase;		// not reg
always @*
	case ({altchar,d[7:6]})
	3'b000: lcase<=0;
	3'b001: lcase<=0;
	3'b010: lcase<=0;
	3'b011: lcase<=1;
	3'b100: lcase<=0;
	3'b101: lcase<=1;
	3'b110: lcase<=0;
	3'b111: lcase<=1;
	endcase

reg invtxt;		// not reg
always @*
	case ({altchar,grd[7:6]})
	3'b000: invtxt<=1;
	3'b001: invtxt<=osc[4];
	3'b010: invtxt<=0;
	3'b011: invtxt<=0;
	3'b100: invtxt<=1;
	3'b101: invtxt<=1;
	3'b110: invtxt<=0;
	3'b111: invtxt<=0;
	endcase

wire [6:0]charout;
wire [9:0]charaddr={lcase,d[5:0],vc[3:1]};

genrom #(
	.INITFILE("character.hex"),
	.AW(10),
	.DW(7)
	) charrom0 (.clk(clk), .addr(charaddr), .data_out(charout));

wire [6:0]txtout = invtxt ? (~charout[6:0]) : charout[6:0];

///// graphics reg / attrib
reg [7:0]grd;
always @(posedge clk) if (rd) grd<=d;

/////////////// Shift register //////////////////
wire [6:0]pixd=gr_en?grd:txtout; // text/hgr

wire ld1=(pixcnt==1) & de & (~clk);
wire ld2=(pixcnt==5) & de & clk &dres;

reg [6:0]sh;
reg rd7=0;

always @(posedge clk2x) begin
	sh<=(ld1|ld2)?pixd:(dres|(~clk)?{1'b0,sh[6:1]}:sh);
end

always @(posedge clk2x) if (ld1) rd7<=(~text)&(~dres)&grd[7];

// half pixel delay
reg dpix;
always @(posedge clk2x) dpix<=sh[0];

wire hrpixel=rd7 ? dpix : sh[0];

////// Lo-RES video //////

reg lrload=0;
always @(posedge clk2x) lrload<=(pixcnt==1)&(clk);

reg [3:0]lrsh;	// shift reg
wire [3:0]lrdata = de ? (vc[3]?grd[7:4]:grd[3:0]):0;
//wire [3:0]lrdata = de ? 8:0;
always @(posedge clk2x) lrsh<=lrload ? lrdata : {lrsh[0],lrsh[3:1]};

reg lrdchr=0;	// delayed character odd/even
always @(posedge clk) if (pixcnt==1) lrdchr<=chrcnt[0];

wire lrpixel=lrdchr?lrsh[2]:lrsh[0];

//reg pdel;
//always @(posedge clk2x) pdel<=lrpixel;
///// Video output /////

wire pixel=(text|hires|(mix&l160))?hrpixel:lrpixel;


///// NTSC Color subcarrier Emulation ////////

wire color=(colorsel==0);

reg [1:0]cq=0; // Subcarrier generator (Johnson counter)
always @(posedge clk2x) cq<=nl?2'b00:{~cq[0],cq[1]};
wire cp0= cq[0]; // Carrier (phase= 0)
wire cp1=~cq[1]; // Carrier delayed #1 sample  (phase= 90)
wire cp2=~cq[0]; // Carrier delayed #2 samples (phase=180)
wire cp3= cq[1]; // Carrier delayed #3 samples (phase=270)

reg [2:0]pix; // old pixels
always @(posedge clk2x) pix<={pix[1:0],pixel};

wire [3:0]cpix=(text|(~color))?0:{pix,pixel}; // pixels for color components

// I color component
wire [2:0]I=(cp3^cpix[0])+(cp0^cpix[1])+(cp1^cpix[2])+(cp2^cpix[3]);
// Q color component
//wire [2:0]Q=(hires^cp0^cpix[0])+(hires^cp1^cpix[1])+(hires^cp2^cpix[2])+(hires^cp3^cpix[3]);
wire [2:0]Q=(cp0^cpix[0])+(cp1^cpix[1])+(cp2^cpix[2])+(cp3^cpix[3]);
//////////////// color ROM (YIQ to RGB) /////////////////////
wire [11:0]avideo; // apple 2 video
genrom #(
	.INITFILE("iqcolor.hex"),
	.AW(7),
	.DW(12)
	) iqcolorrom0 (.clk(clk2x), .addr({Y,I,Q}), .data_out(avideo));

wire Y=pix[1];
wire [3:0]R=avideo[11:8];
wire [3:0]G=avideo[7:4];
wire [3:0]B=avideo[3:0];
	
// green monochrome monitor and virtual leds
wire [11:0]cmask=((colorsel==2'b01) ? GREENCOLOR: 12'b1111_1111_1111)&
				 ((colorsel==2'b10) ? AMBERCOLOR: 12'b1111_1111_1111);
assign video= (avideo&cmask) | vcolor;

// virtual leds for debug 
`ifdef SOFTCARD
parameter NLEDS=5;
`else
parameter NLEDS=4;
`endif
wire [7:0]ledcolor=LEDCOLOR;
wire vonrange=(vc[8:4]<NLEDS); // 5 leds
wire dbgon=(chrcnt==56)&vonrange&vde&(pixcnt!=0)&(pixcnt!=6)&(vc[3:0]>3)&(vc[3:0]<11);
wire dbgmark=(chrcnt==56)&vonrange&vde&(vc[3:0]==7);
wire dbgpix=((dbgon&vleds[vc[6:4]])|dbgmark)&enleds;
wire [11:0]vcolor=dbgpix? (ledcolor[vc[6:4]]?12'hFb0:12'hF00) : 0;


endmodule

