///////////////////////////////////////////////////////
// Apple 2 Floppy Disk controller emulation
//  by Jesus Arias (2023)
//  

`define BRAM_PROM6

module wozmachine (
	input clk,		// CPU clk (1MHz)
	input clk2x,	// double clock speed (2MHz)
	input reset,	// initial state
	input we,		// write if 1
	input cs,		// card select
	input [3:0]a,	// address bus
	input [7:0]d,	// data input
	output [7:0]q,	// data output
	
	output [3:0]phs,	// stepper motor phases
	output drvon,	// motor ON
	output drvsel,	// disk 1 / disk 2
	input  wrprot,	// write protection input
	output wrrq,	// write request
	output wrdata,	// Write data stream	
	input  rddata	// Read data stream
	
);

/////////////////////////////////////////////////////////////
///////////////////////// Controller ////////////////////////
/////////////////////////////////////////////////////////////

///////// 74LS259 addresable latch /////////
reg [7:0]alat=0;
always @(posedge clk or posedge reset)
	if (reset) alat<=0;
	else if (cs) alat[a[3:1]]<=a[0];
assign phs=alat[3:0];
assign drvsel = alat[5];
wire shload = alat[6];
wire rdwr   = alat[7];
assign wrrq = alat[7];

wire rdc = (a[3:1]==3'b110) & cs; //read data strobe (debug)

//////////////////// Drive off delay (1sec) /////////////////
reg [19:0]del555=0;
wire motoron  = (del555!=0);

always @(posedge clk) 
	if (alat[4]) del555<=20'hfffff;
	else if (motoron) del555<=del555-1;

assign drvon = motoron; 

/////////////////// Wozniak's state machine /////////////////

//////// 74LS174 (hex FF) /////////
reg [5:0]dff=0;

always @(negedge clk2x or negedge drvon)
	if (~drvon) dff<=0; else dff<={rddata,dff[5],pq[4],pq[5],pq[6],pq[7]};
wire fallingrd = ~((~dff[5])&dff[4]);	// falling read pulse
wire wrdata = dff[0]&rdwr;

/////// 74LS323 shift register ///////
wire srs0=pq[1];
wire srs1=pq[0];
wire srclr=pq[3];
wire srdsl=pq[2];
wire srqa=sr[7];

reg [7:0]sr=0;
always @(negedge clk2x)
	if (~srclr) sr<=0;
	else case ({srs1,srs0})
		2'b00: sr<=sr;					// NOP
		2'b01: sr<={wrprot,sr[7:1]};	// SHR
		2'b10: sr<={sr[6:0],srdsl};		// SHL
		2'b11: sr<=d;					// LOAD
	endcase

assign q=sr;	// reads from $C0Ex

// State machine ROM

wire [7:0]pa={dff[0],dff[1],dff[3],fallingrd,rdwr,shload,srqa,dff[2]};

`ifdef BRAM_PROM6
// BRAM based ROM (synchronous)
genrom #(
	.INITFILE("prom6.hex"),
	.AW(8),
	.DW(8)
	) prom6 (.clk(clk2x), .addr(pa), .data_out(pq));
wire [7:0]pq;
`else
// Logic cell based ROM (asynchronous)
reg [7:0]prom6[0:255];
initial begin
  $readmemh("prom6.hex", prom6);
end
wire [7:0]pq=prom6[pa];
`endif

endmodule
